************************************************************************
*	Generic amplitude program for arbitrary processes involving
*	8 external fermion legs
*
*       Initialization by 		subroutine initparam
*       Squared matrix element by 	function M2(p)
*
*	Checks: amplitudes for e+ e- --> 6f checked against Madgraph
*		for 6f states with at most 3 neutrinos and at most
*		4 quarks, with the exception of ee->eeeeee which was
*		not feasible with Madgraph
*
************************************************************************
        subroutine initparam
************************************************************************
*	initialize couplings etc. for matrix element with incoming
*	f1bar f2 f3bar f4 f5bar f6 f7bar f8
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 ccw,csw,cmw,cmt
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 qi(20,8),nspec(8),nspecfact(8),sym(20),wgt(20)
	integer i,j,i1,i2,f,f1,f2,v,v1,v2,v3,v4,npol,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	character*6 fname(20,8),dfname(20,8)

	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
        common/lparam/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,gh,me,me2,mt,mt2,gt

	do 99 fs=1,nfs

C***	Check for lepton- and quark-number conservation
	if(quark(fs,1)-quark(fs,2)+quark(fs,3)-quark(fs,4)
     &	  +quark(fs,5)-quark(fs,6)+quark(fs,7)-quark(fs,8).ne.0) then
	  write(*,*) 'Process no. ',fs
	  write(*,*) 'No lepton- and quark-number conservation !'
	  stop
	endif

C***	Check for weak-isospin conservation
	if(iso(fs,1)-iso(fs,2)+iso(fs,3)-iso(fs,4)
     &	  +iso(fs,5)-iso(fs,6)+iso(fs,7)-iso(fs,8).ne.0) then
	  write(*,*) 'Process no. ',fs
	  write(*,*) 'No weak-isospin conservation !'
	  stop
	endif

C***	Check for conservation of generation indices
	if(gen(fs,1)-gen(fs,2)+gen(fs,3)-gen(fs,4)
     &	  +gen(fs,5)-gen(fs,6)+gen(fs,7)-gen(fs,8).ne.0) then
	  write(*,*) 'Process no. ',fs
	  write(*,*) 'No generation conservation !'
	  stop
	endif

c*** 	Check for external top quarks
	do i=1,8
	if ((quark(fs,i).eq.1).and.(iso(fs,i).eq.1).and.
     &	    (gen(fs,i).eq.3)) then
	  write(*,*) 'Process no. ',fs
	  write(*,*) 'No external top quarks allowed !'	
	  stop
	endif
	enddo

cc*** 	Check for more than 4 external quarks
c	if (quark(fs,1)+quark(fs,2)+quark(fs,3)+quark(fs,4)+
c     &	    quark(fs,5)+quark(fs,6)+quark(fs,7)+quark(fs,8).gt.4) then
c	  write(*,*) 'Process no. ',fs
c	  write(*,*) 'Not more than 4 external quarks allowed !'	
c	  stop
c	endif

99	continue

C*** 	Check for 2->6 reaction
	if(dir(1)+dir(2)+dir(3)+dir(4)
     &	  +dir(5)+dir(6)+dir(7)+dir(8).ne.-4) then
	  write(*,*) 'Process is not 2 -> 6 !'
	  stop
	endif

C*** 	Masses and widths for propagators 
	mp(0)   = mh
	mp(1)   = 0d0
	mp(2)   = mz
	mp(3)   = mw
	mp(4)   = mt
	gp(0)   = gh
	gp(1)   = 0d0
	gp(2)   = gz
	gp(3)   = gw
	gp(4)   = gt
	do 100 v=0,4
	  mp2( v)= mp(v)**2
	  mp2(-v)= mp2(v)
          mp(-v) = mp(v)
	  gp(-v) = gp(v)
100	continue

C*** complex or real parameters for couplings
	if (qprop.eq.4) then
	  ccw = sqrt( dcmplx(mw**2,-mw*gw)/dcmplx(mz**2,-mz*gz) )
	  csw = sqrt( 1d0-ccw**2 )
	  cmw = sqrt( dcmplx(mw**2,-mw*gw) )
	  cmt = sqrt( dcmplx(mt**2,-mt*gt) )
	else
	  ccw = cw
	  csw = sw
	  cmw = mw
	  cmt = mt
	endif 
	mf(1) = 0d0
	mf(4) = cmt
	gs2   = 4d0*pi*alphas
	
	do 199 fs=1,nfs

C***	Elmg. charges of fermions
	do 200 f=1,8
	  q(fs,f)  = (iso(fs,f)-1d0)/2d0+quark(fs,f)*2d0/3d0
	  qi(fs,f) = q(fs,f)-iso(fs,f)
200	continue

C*** 	Couplings of fermions to bosons 
	do 300 f=1,8
	  Cvf(fs,-1,1,f) = -el*q(fs,f)
	  Cvf(fs,+1,1,f) = Cvf(fs,-1,1,f)
	  Cvf(fs,-1,2,f) = -el*(q(fs,f)*csw/ccw-iso(fs,f)/2d0/csw/ccw)
	  Cvf(fs,+1,2,f) = -el* q(fs,f)*csw/ccw
	  Cvf(fs,-1,+iso(fs,f)*(-1)**f*3,f) = el/sqrt(2d0)/csw
	  Cvf(fs,+1,-iso(fs,f)*(-1)**f*3,f) = 0d0
	  Cvif(fs,-1,1,f) = -el*qi(fs,f)
	  Cvif(fs,+1,1,f) = Cvif(fs,-1,1,f)
	  Cvif(fs,-1,2,f) = -el*(qi(fs,f)*csw/ccw+iso(fs,f)/2d0/csw/ccw)
	  Cvif(fs,+1,2,f) = -el* qi(fs,f)*csw/ccw
	  Cvif(fs,-1,-iso(fs,f)*(-1)**f*3,f) = el/sqrt(2d0)/csw
	  Cvif(fs,+1,+iso(fs,f)*(-1)**f*3,f) = 0d0
	  do i=-1,1,2
	    do j=1,2
	      Cvf(fs,i,-j,f)  = Cvf(fs,i,j,f)
	      Cvif(fs,i,-j,f) = Cvif(fs,i,j,f)
	    enddo
	  enddo
	  fv(fs,f, 0) = 1
	  fv(fs,f,+1) = 1
	  fv(fs,f,+2) = 1
	  fv(fs,f,+3) = 1
	  fv(fs,f,-3) = 1
	  if ((quark(fs,f).eq.1).and.(gen(fs,f).eq.3).and.
     &        (iso(fs,f).eq.-1))       fv(fs,f,-(-1)**f*3) = 4
300	continue
	do 400 i1=-1,1,2
	do 400 i2=-1,1,2
	do 400 f1=1,8
	do 400 f2=1,8
	do 400 v1=-3,3
	do 400 v2=-3,3
	  Cvfvf(fs,i1,v1,f1,i2,v2,f2) =Cvf(fs,i1,v1,f1)*Cvf(fs,i2,v2,f2)
400	continue
199	continue

c***    Top-quark Yukawa coupling
	Csbt = el/sqrt(2d0)/csw*cmt/cmw

C***	Triple and quartic gauge-boson couplings
	do 500 v1=-3,3
	do 500 v2=-3,3
	do 500 v3=-3,3
	  Cvvv(v1,v2,v3) = 0d0
	  Cvvs(v1,v2,v3) = 0d0
	do 500 v4=-3,3
	  Cwwvv(v1,v2,v3,v4) = 0d0
500	continue
	Cvvv( 1,-3, 3)  =  el
	Cvvv( 2,-3, 3)  = -el*ccw/csw
	Cvvs( 0, 3,-3)  =  0d0
	Cvvs( 1, 3,-3)  = -el*cmw
	Cvvs( 2, 3,-3)  = -el*cmw*csw/ccw
	Cwwvv(-3,-3,3,3)=  el**2/csw**2
	Cwwvv(3,3,-3,-3)=  Cwwvv(-3,-3,3,3)
	Cvvs( 2, 2, 0)  =  el*cmw/csw/ccw**2
	Cvvs( 2,-2, 0)  =  Cvvs(2,2,0)
	Cvvs( 3,-3, 0)  =  el*cmw/csw
	Cvvs(-3, 3, 0)  =  Cvvs(3,-3,0)
	do j=1,2
	  Cvvv(-j,-3, 3) = Cvvv( j,-3, 3)
	  Cvvs(-j, 3,-3) = Cvvs( j, 3,-3)
	  Cvvs(-j,-j, 0) = Cvvs( j, j, 0)
	  Cvvs(-j, j, 0) = Cvvs( j, j, 0)
	enddo
	do j=-2,2
	  Cvvv(-3, 3, j) =  Cvvv( j,-3, 3)
	  Cvvv( 3, j,-3) =  Cvvv( j,-3, 3)
	  Cvvv( j, 3,-3) = -Cvvv( j,-3, 3)
	  Cvvv( 3,-3, j) = -Cvvv( j,-3, 3)
	  Cvvv(-3, j, 3) = -Cvvv( j,-3, 3)
	  Cvvs( 3, j,-3) =  Cvvs( j, 3,-3)
	  Cvvs( j,-3, 3) =  Cvvs( j, 3,-3)
	  Cvvs(-3, j, 3) =  Cvvs( j, 3,-3)
	enddo
	do 600 v3=-2,2
	do 600 v4=-2,2
	  Cwwvv(-3,3,v3,v4) = -Cvvv(v3,-3,3)*Cvvv(v4,-3,3)
600	continue

	do 601 fs=1,nfs
c*** Symmetry factor for identical particles
	sym(fs) = 1d0
	do 610 i1=1,8
	  nspec(i1)     = 1d0
	  nspecfact(i1) = 1d0
	do 620 i2=1,8
	  if((i1.eq.i2).or.(mod(i1-i2,2).ne.0)) goto 620
	  if((quark(fs,i1).eq.quark(fs,i2)).and.
     &	     (iso(fs,i1).eq.iso(fs,i2)).and.
     &	     (gen(fs,i1).eq.gen(fs,i2)).and.(dir(i1).eq.dir(i2))) then
      	    nspec(i1)     = nspec(i1) + 1d0
      	    nspecfact(i1) = nspecfact(i1)*nspec(i1)
	  endif
620	continue
	  sym(fs) = sym(fs)/nspecfact(i1)**(1d0/nspec(i1))
610	continue

C*** 	Weights for polarization channels
        do 700 i1=-1,1,2
          f1p  = 1d0+pp*i1
        do 700 i2=-1,1,2
          f2p  = 1d0+pm*i2
        do 700 i3=-1,1,2
          g3p  = (2+pol(3)*i3)/2
        do 700 i4=-1,1,2
          g4p  = (2+pol(4)*i4)/2
        do 700 i5=-1,1,2
          g5p  = (2+pol(5)*i5)/2
        do 700 i6=-1,1,2
          g6p  = (2+pol(6)*i6)/2
        do 700 i7=-1,1,2
          g7p  = (2+pol(7)*i7)/2
        do 700 i8=-1,1,2
          g8p  = (2+pol(8)*i8)/2
          npol = (255+i1+2*i2+4*i3+8*i4+16*i5+32*i6+64*i7+128*i8)/2
          polw(fs,npol) = f1p*f2p*g3p*g4p*g5p*g6p*g7p*g8p
     &		          /4d0*sym(fs)*wgt(fs)
700     continue
601	continue

C*** 	Initialize generic treatment of amplitudes
	call initgeneric

	end

************************************************************************
        subroutine initgeneric
************************************************************************
*	initialize generic treatment of amplitudes
*-----------------------------------------------------------------------
*       8.8.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 qi(20,8),sym(20),wgt(20)
	real*8 colint(20,24,24),colintgg(20,24,24),colintg(20,24,24)
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer i,j,k,l,m,i1,i2,i3,i4,i5,i6,i7,i8,eps(4,4,4,4)
	integer sgn1,sgn2,fs
	integer nfperm(20),fperm(20,576,8),qvperm(20,576,4)
	integer fpermsgn(20,576)
	integer qlperm(20,576,0:4),ncclass(20),cclass(20,24,0:4)
	integer coldia(20),colav(20),qlcycle(20,4,8),n(0:8)
	character*6 fname(20,8),dfname(20,8)

	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/generic/nfperm,fperm,qvperm,fpermsgn,qlperm
	common/colour/colint,colintgg,colintg,
     &		ncclass,cclass,coldia,colav

	do 300 i1=1,4
	do 300 i2=1,4
	do 300 i3=1,4
	do 300 i4=1,4
	  eps(i1,i2,i3,i4) = 0
300	continue
        EPS(1,2,3,4) =  1
        EPS(1,2,4,3) = -1
        EPS(1,3,2,4) = -1
        EPS(1,3,4,2) =  1
        EPS(1,4,2,3) =  1
        EPS(1,4,3,2) = -1
        EPS(2,1,3,4) = -1
        EPS(2,1,4,3) =  1
        EPS(2,3,1,4) =  1
        EPS(2,3,4,1) = -1
        EPS(2,4,1,3) = -1
        EPS(2,4,3,1) =  1
        EPS(3,1,2,4) =  1
        EPS(3,1,4,2) = -1
        EPS(3,2,1,4) = -1
        EPS(3,2,4,1) =  1
        EPS(3,4,1,2) =  1
        EPS(3,4,2,1) = -1
        EPS(4,1,2,3) = -1
        EPS(4,1,3,2) =  1
        EPS(4,2,1,3) =  1
        EPS(4,2,3,1) = -1
        EPS(4,3,1,2) = -1
        EPS(4,3,2,1) =  1

c*** fermion permutations for generic graphs:  fperm(k,i)    
c		k=1,...,nfperm  counts permutations
c		i=1,...,8       counts f's in permutation k
c
c	f-fbar pairs for generic graphs:
c		fa = fperm(k,1)  ---  fb(bar) = fperm(k,2)
c		fc = fperm(k,3)  ---  fd(bar) = fperm(k,4)
c		fe = fperm(k,5)  ---  ff(bar) = fperm(k,6)
c		ff = fperm(k,7)  ---  fh(bar) = fperm(k,8)
c
c	permutations of anti-quark lines w.r.t. quark lines:  
c		qlperm(k,i)	end position of line i
c				where k=1,...,nfperm counts permutations
c				and i=1,...,4       
c
c		qlperm(k,0)	number k=1,...,ncclass of colour class
c
c		cclass(j,i)	permutation (e.g. '1243') of
c				colour class j with j=1,...,ncclass
c				with i=1,2,3,4 as in qlperm(k,i)
c				and for i=0 all i=1,2,3,4 combined

	do 99 fs=1,nfs

	nfperm(fs) =0
	ncclass(fs)=0
	do 100 i1=1,7,2
	do 100 i3=1,7,2
	do 100 i5=1,7,2
	do 100 i7=1,7,2
	sgn1 = eps((i1+1)/2,(i3+1)/2,(i5+1)/2,(i7+1)/2)
	if (sgn1.eq.0) goto 100
	do 200 i2=2,8,2
	do 200 i4=2,8,2
	do 200 i6=2,8,2
	do 200 i8=2,8,2
	sgn2 = eps(i2/2,i4/2,i6/2,i8/2)
	if (sgn2.eq.0) goto 200
c*** check fermion-antifermion pairs
	if((gen(fs,i1).ne.gen(fs,i2)).or.(gen(fs,i3).ne.gen(fs,i4)).or.
     &	   (gen(fs,i5).ne.gen(fs,i6)).or.(gen(fs,i7).ne.gen(fs,i8))) 
     &		goto 200
	if((quark(fs,i1).ne.quark(fs,i2)).or.
     &	   (quark(fs,i3).ne.quark(fs,i4)).or.
     &	   (quark(fs,i5).ne.quark(fs,i6)).or.
     &	   (quark(fs,i7).ne.quark(fs,i8)))	goto 200
c*** fermion permutation
	  nfperm(fs) = nfperm(fs)+1
	  fperm(fs,nfperm(fs),1) = i1
	  fperm(fs,nfperm(fs),2) = i2
	  fperm(fs,nfperm(fs),3) = i3
	  fperm(fs,nfperm(fs),4) = i4
	  fperm(fs,nfperm(fs),5) = i5
	  fperm(fs,nfperm(fs),6) = i6
	  fperm(fs,nfperm(fs),7) = i7
	  fperm(fs,nfperm(fs),8) = i8
	  fpermsgn(fs,nfperm(fs))= sgn1*sgn2
	  do i=1,8
	    n(fperm(fs,nfperm(fs),i)) = i
	  enddo
c*** quark line permutation
	  qlperm(fs,nfperm(fs),0) = 0
	  do i=1,4
	    qlperm(fs,nfperm(fs),i) =(fperm(fs,nfperm(fs),n(2*i)-1)+1)/2
     &			        *quark(fs,fperm(fs,nfperm(fs),n(2*i)-1))
	    qlperm(fs,nfperm(fs),0) = qlperm(fs,nfperm(fs),0) 
     &				+ 10**(4-i)*qlperm(fs,nfperm(fs),i)
	  enddo
	  j=1
101	  continue
	  if (j.gt.ncclass(fs)) then
	  cclass(fs,j,0) = 0
	    do i=1,4
	    cclass(fs,j,i) = qlperm(fs,nfperm(fs),i)
	    cclass(fs,j,0) = cclass(fs,j,0) + 10**(4-i)*cclass(fs,j,i)
	    enddo
	    ncclass(fs)   = ncclass(fs)+1
	    goto 102
	  endif
	  if(qlperm(fs,nfperm(fs),0).eq.cclass(fs,j,0)) goto 102
	  j = j+1
	  goto 101
102	  continue
	  qlperm(fs,nfperm(fs),0) = j

c*** incoming charges
	  qvperm(fs,nfperm(fs),1) = int((q(fs,i2)-q(fs,i1))*1.00001d0)
	  qvperm(fs,nfperm(fs),2) = int((q(fs,i4)-q(fs,i3))*1.00001d0)
	  qvperm(fs,nfperm(fs),3) = int((q(fs,i6)-q(fs,i5))*1.00001d0)
	  qvperm(fs,nfperm(fs),4) = int((q(fs,i8)-q(fs,i7))*1.00001d0)
c	write(*,*) nfperm,i1,i2,i3,i4,i5,i6,i7,i8,
c     &		   qvperm(nfperm,1),qvperm(nfperm,2),
c     &		   qvperm(nfperm,3),qvperm(nfperm,4)
200	continue
100	continue

c*** colour factors 
c    colour average
	colav(fs)  = 1
	do i=1,8
	  if ((quark(fs,i).eq.1).and.(dir(i).eq.1)) 
     &		colav(fs) = 3*colav(fs)
	enddo
c    colour multiplicity for diagonal terms
        coldia(fs) = 1d0
        do i=1,7,2
          if (quark(fs,i).eq.1) coldia(fs) = 3*coldia(fs)
        enddo

c    colour multiplicities for interference terms --> classes
c    based on line permutations
c    colint(i,j) -> purely electroweak diagrams squared
	do 1400 i=1,ncclass(fs)
	do 1400 j=1,i
	  do l=1,4
	    n(cclass(fs,j,l)) = l
	  enddo
          cint = 1d0
	do 1600 k=1,4
	  do l=1,8
	    qlcycle(fs,k,l) = 0
	  enddo
	  l = 1
          if (cclass(fs,i,k).ne.0) then
	    qlcycle(fs,k,1) = k
1500	    continue
	    l = l+1
	    qlcycle(fs,k,l) = cclass(fs,i,qlcycle(fs,k,l-1))
	    if (n(qlcycle(fs,k,l)).eq.qlcycle(fs,k,1)) goto 1501
	    l = l+1
	    qlcycle(fs,k,l) = n(qlcycle(fs,k,l-1))
	    goto 1500
1501	    continue
            cint = cint*3**(2d0/dfloat(l))
	  endif
1600	continue
        colint(fs,i,j) = int(cint+1d-10)
        colint(fs,j,i) = colint(fs,i,j)
1400	continue

c    colintgg(i,j) -> one-gluon exchange diagrams squared
	do 1700 i=1,ncclass(fs)
	do 1700 j=1,i
	  if (colint(fs,i,j).eq.9d0) then
	    colintgg(fs,i,j) = 2d0
	    colintg(fs,i,j)  = 0d0
	  elseif (colint(fs,i,j).eq.3d0) then
	    colintgg(fs,i,j) = -2d0/3d0
	    colintg(fs,i,j)  = 4d0
	  endif
          colintgg(fs,j,i) = colintgg(fs,i,j)
          colintg(fs,j,i)  = colintg(fs,i,j)
1700	continue

99	continue

	end

************************************************************************
        subroutine initperm(fs,nfp)
************************************************************************
*	initialize nfp-th fermion permutation (fa,fb,fc,fd,fe,ff,fg,fh)
*	of fs-th final state
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	real*8 colint(20,24,24),colintgg(20,24,24),colintg(20,24,24)
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,nfp,fs
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer nfperm(20),fperm(20,576,8),qvperm(20,576,4)
	integer qlperm(20,576,0:4),ncclass(20),cclass(20,24,0:4)
	integer coldia(20),colav(20),fpermsgn(20,576)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/generic/nfperm,fperm,qvperm,fpermsgn,qlperm
	common/colour/colint,colintgg,colintg,
     &		ncclass,cclass,coldia,colav

	if (nfp.gt.nfperm(fs)) then
	  write(*,*) 'Number of fermion permutations out of range !'
	  stop
	endif

	fa = fperm(fs,nfp,1)
	fb = fperm(fs,nfp,2)
	fc = fperm(fs,nfp,3)
	fd = fperm(fs,nfp,4)
	fe = fperm(fs,nfp,5)
	ff = fperm(fs,nfp,6)
	fg = fperm(fs,nfp,7)
	fh = fperm(fs,nfp,8)

	qab = qvperm(fs,nfp,1)
	qcd = qvperm(fs,nfp,2)
	qef = qvperm(fs,nfp,3)
	qgh = qvperm(fs,nfp,4)

	sgn = fpermsgn(fs,nfp)

	ccl = qlperm(fs,nfp,0)

	n(fa) = 1
	n(fb) = 2
	n(fc) = 3
	n(fd) = 4
	n(fe) = 5
	n(ff) = 6
	n(fg) = 7
	n(fh) = 8

	yab(+1) = fa	
	yab(-1) = fb
	ycd(+1) = fc
	ycd(-1) = fd
	yef(+1) = fe	
	yef(-1) = ff
	ygh(+1) = fg
	ygh(-1) = fh

	end

************************************************************************
        subroutine setprods(p,dir)
************************************************************************
* 	Weyl-van der Waerden and Minkowski products
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 eiph(8),sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4)
	complex*16 prop
	real*8 e(8),th(8),re(8),cth2(8),sth2(8),vp(8,8),p(8,0:3)
	integer i,j,k,dir(8)

	common/prods/sp,csp,bsp,vp,propvp

        do 100 i=1,8
          e(i)    = p(i,0)
	  re(i)   = sqrt(e(i))
          vec     = sqrt(p(i,1)**2+p(i,2)**2+p(i,3)**2)
          th(i)   = acos(p(i,3)/vec)
	  cth2(i) = cos(th(i)/2d0) 
	  sth2(i) = sin(th(i)/2d0) 
          if (abs(sin(th(i))).lt.1d-10) then
            eiph(i) = 1d0
          else
            eiph(i) = dcmplx(p(i,1),p(i,2))/vec/sin(th(i))
          endif
100     continue

	do 200 i=2,8
	do 200 j=1,i-1
	   sp(i,j) = 2d0*re(i)*re(j)
     &	             *(cth2(i)/eiph(i)*sth2(j)-sth2(i)*cth2(j)/eiph(j))
	  csp(i,j) = dconjg(sp(i,j))*dir(i)*dir(j)
	   vp(i,j) = sp(i,j)*csp(i,j)
	   sp(j,i) = - sp(i,j)
	  csp(j,i) = -csp(i,j)
	   vp(j,i) =   vp(i,j)
	do 200 k=0,3
	  propvp(i,j, k) = Prop(vp(i,j),k)
	  propvp(j,i, k) = propvp(i,j,k)
	  propvp(i,j,-k) = propvp(i,j,k)
	  propvp(j,i,-k) = propvp(i,j,k)
200	continue

	do 300 i=1,8
	do 300 k=1,8
	do 300 j=1,8
	  bsp(i,k,j) = csp(i,k)*sp(j,k)
300	continue

	end

************************************************************************
        function Prop(p2,i)
************************************************************************
*	Generic gauge-boson propagator
*-----------------------------------------------------------------------
*	general form:  1 / ( p2 - mv(p2) ),    
*	      bosons:  j = h,a,z,w,t
*	  qprop = 0:  mv(p2) = mv(j)^2,              
*	  qprop = 1:  mv(p2) = mv(j)^2 - i*mv(j)*gv(j),              
*	  qprop = 2:  mv(p2) = mv(j)^2 - i*mv(j)*gv(j)*theta(p2),     
*	  qprop = 3:  mv(p2) = mv(j)^2 - i*p2/mv*gv(j)*theta(p2),  
*-----------------------------------------------------------------------
*       25.8.98 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 Prop
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 qi(20,8),sym(20),wgt(20)
        integer i
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	character*6 fname(20,8),dfname(20,8)

	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname

	theta(x) = .5d0+sign(.5d0,x)

	if ((qprop.eq.0).or.(i.eq.1)) then
          Prop = 1d0/( p2-mp2(i) )
          return
	elseif ((qprop.eq.1).or.(qprop.eq.4)) then
          Prop = 1d0/( p2-mp2(i)+dcmplx(0d0,mp(i)*gp(i)) )
          return
	elseif (qprop.eq.2) then
          Prop = 1d0/( p2-mp2(i)+dcmplx(0d0,mp(i)*gp(i))*theta(p2) )
          return
	elseif (qprop.eq.3) then
          Prop = 1d0/( p2-mp2(i)
     &		        +dcmplx(0d0,p2/mp(i)*gp(i))*theta(p2) )
          return
        endif

	write(*,*) 'Propagator flag not defined, qprop = ',qprop
	stop

	end

************************************************************************
        subroutine setMnull(M)
************************************************************************
*	initialize matrix element M
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
	complex*16 M(0:255,24)
	integer    i,j

	do i=0,255
	do j=1,24
	  M(i,j) = 0d0
	enddo
	enddo

	end

************************************************************************
        function M2(p)
************************************************************************
*	calculation of |M|^2
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),Mg(0:255,24)
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
        real*8 p(8,0:3),p1(0:3,8),ans(1),sym(20),wgt(20),qi(20,8)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 colint(20,24,24),colintgg(20,24,24),colintg(20,24,24)
	real*8 m2fs(20)
        integer npol,i1,i2,i3,i4,i5,i6,i7,i8,qa,qz,qg,i
        integer g3p,g4p,g5p,g6p,g7p,g8p
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs,fpermsgn(20,576)
	integer nfperm(20),fperm(20,576,8),qvperm(20,576,4)
	integer qlperm(20,576,0:4),ncclass(20),cclass(20,24,0:4)
	integer coldia(20),colav(20)
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer ccl1,ccl2,fs
	character*6 fname(20,8),dfname(20,8)

	common/generic/nfperm,fperm,qvperm,fpermsgn,qlperm
	common/colour/colint,colintgg,colintg,
     &		ncclass,cclass,coldia,colav
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
        common/lparam/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,gh,me,me2,mt,mt2,gt
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
        common/qcdoption/qg
        common/mat2/m2fs

	n4a = 0
	n4b = 0
	n4c = 0
	n3a = 0
	n3b = 0
	n3c = 0
	n3d = 0
	n3e = 0
	n2a = 0
	n2b = 0
	n2c = 0
	n2d = 0

	qa = 1
	qz = 1
c	qg = 1

	M2 = 0d0
	do 100 fs=1,nfs

        call setMnull(M)
        call setMnull(Mg)
	do 200 i=1,nfperm(fs)
          call initperm(fs,i)

c*** purely electroweak diagrams
  	  call addM4a(M,fs,qa,qz,qa,qz)
 	  call addM4b(M,fs,qa,qz,qa,qz)
 	  call addM4c(M,fs,qa,qz,qa,qz,qa,qz)
 	  call addM3a(M,fs,qa,qz,0,qa,qz,0,qa,qz,0)
 	  call addM3b(M,fs,qa,qz)
 	  call addM3c(M,fs,qa,qz)
	  call addM3d(M,fs,qa,qz,0,qa,qz)
	  call addM3e(M,fs,qa,qz,0,qa,qz)
 	  call addM2a(M,fs)
 	  call addM2b(M,fs,qa,qz,0,qa,qz,0,qa,qz,0)
 	  call addM2c(M,fs,qa,qz,0,qa,qz,0,qa,qz,0)
   	  call addM2d(M,fs,qa,qz,0,qa,qz,0,qa,qz,0)

c*** diagrams with one-gluon exchange
c    only consistent for exactly four external quarks!
	  if (qg.eq.0) goto 200
	  if (quark(fs,fa)*quark(fs,fc).eq.1) then
	    call addM3a(Mg,fs, 0, 0,qg,qa,qz,0,qa,qz,0)
	    call addM3d(Mg,fs, 0, 0,qg,qa,qz)
	    call addM3e(Mg,fs, 0, 0,qg,qa,qz)
 	    call addM2b(Mg,fs,qa,qz,0, 0, 0,qg,qa,qz,0)
 	    call addM2c(Mg,fs,qa,qz,0, 0, 0,qg,qa,qz,0)
 	    call addM2d(Mg,fs,qa,qz,0, 0, 0,qg,qa,qz,0)
	  elseif (quark(fs,fa)*quark(fs,fe).eq.1) then
	    call addM3a(Mg,fs,qa,qz,0, 0, 0,qg,qa,qz,0)
	    call addM2b(Mg,fs, 0, 0,qg,qa,qz,0,qa,qz,0)
 	    call addM2c(Mg,fs, 0, 0,qg,qa,qz,0,qa,qz,0)
 	    call addM2d(Mg,fs, 0, 0,qg,qa,qz,0,qa,qz,0)
	  elseif (quark(fs,fa)*quark(fs,fg).eq.1) then
	    call addM3a(Mg,fs,qa,qz,0,qa,qz,0, 0, 0,qg)
	  elseif (quark(fs,fc)*quark(fs,fg).eq.1) then
 	    call addM2b(Mg,fs,qa,qz,0,qa,qz,0, 0, 0,qg)
 	    call addM2c(Mg,fs,qa,qz,0,qa,qz,0, 0, 0,qg)
 	    call addM2d(Mg,fs,qa,qz,0,qa,qz,0, 0, 0,qg)
	  endif

200	continue

c*** squaring diagrams
c    square of purely electroweak diagrams
	M2fs(fs) = 0d0
        do npol=0,255
          do ccl1=1,ncclass(fs)
          do ccl2=1,ccl1-1
            M2fs(fs) = M2fs(fs) + 2d0*polw(fs,npol)*colint(fs,ccl1,ccl2)
     &			*dreal(M(npol,ccl1)*dconjg(M(npol,ccl2)))
	  enddo
            M2fs(fs) = M2fs(fs) + polw(fs,npol)*colint(fs,ccl1,ccl1)
     &			   *abs(M(npol,ccl1))**2
	  enddo
	enddo 
c    square of one-gluon diagrams 
        do npol=0,255
          do ccl1=1,ncclass(fs)
          do ccl2=1,ccl1-1
            M2fs(fs) = M2fs(fs)+2d0*polw(fs,npol)*colintgg(fs,ccl1,ccl2)
     &			*dreal(Mg(npol,ccl1)*dconjg(Mg(npol,ccl2)))
	  enddo
            M2fs(fs) = M2fs(fs) + polw(fs,npol)*colintgg(fs,ccl1,ccl1)
     &	              *abs(Mg(npol,ccl1))**2
	  enddo
	enddo 
c    interference of purely electroweak and one-gluon diagrams 
        do npol=0,255
          do ccl1=1,ncclass(fs)
          do ccl2=1,ccl1-1
            M2fs(fs) = M2fs(fs) +2d0*polw(fs,npol)*colintg(fs,ccl1,ccl2)
     &			*dreal( Mg(npol,ccl1)*dconjg( M(npol,ccl2))
     &			       + M(npol,ccl1)*dconjg(Mg(npol,ccl2)))
	  enddo
	  enddo
	enddo 

888	continue

*** comparsion with Madgraph
c	do i1=0,3
c	  do i2=1,8
c	    p1(i1,i2) = p(i2,i1)
c	  enddo
c	enddo
c        call SAMPLITUDE6(P1,ANS)
c        amplitude=ANS(1)
999	format(3g24.16)
c        write(*,999) amplitude/M2fs(fs),M2fs(fs),amplitude

	M2 = M2 + M2fs(fs)
100	continue

	end

************************************************************************
        subroutine addM4a(M,fs,a1,z1,a2,z2)
************************************************************************
*                             \       /
*                          fh _\|   _/ fg
*                               \   /|
*                                \ / 
*                   \             *             /
*                    \_           <           |/_
*                 fa |\           < V2        /  ff
*                      \    W-    <          / 
*                       *vvv > vvv*vvvvvvvvv*  
*                      /          <   V1     \
*                 fb _/         W <          _\| fe
*                    /|           <            \
*                   /             *             \
*                                / \
*                           fc |/_  \_ fd
*                              /    |\ 
*                             /       \
*	
*	Fields: fermions fa...fh incoming 
*	        bosons V flowing out of ffV vertices
*	
*	Insertions: W- W V1 V2 = W- W- W+  W+     with fa<fc, fe<fg
*	                         W- W+ A/Z A/Z    with fe<fg
*-----------------------------------------------------------------------
*       18.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),A(-1:1,-1:1),App,Ce(-1:1),Cg(-1:1),fac
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer a1,z1,a2,z2,xe,xf,xg,xh,i(8),i5,i7,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol,j
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

 	App(xe,xf,xg,xh) = 2d0*csp(fb,fd)*csp(xf,xh)*sp(fa,fc)*sp(xe,xg)
     &	                      -csp(fb,xf)*csp(fd,xh)*sp(fa,xe)*sp(fc,xg)
     &	                      -csp(fb,xh)*csp(fd,xf)*sp(fa,xg)*sp(fc,xe)

	if((.not.((qab.eq.-1).and.(qcd.eq.-1).and.(qef.eq.1).and.
     &	          (fa.lt.fc).and.(fe.lt.fg))).and.
     &	   (.not.((qab.eq.-1).and.(qcd.eq.+1).and.(qef.eq.0).and.
     &	          (fe.lt.fg)))) return

	fac = sgn*4d0*Cvfvf(fs,-1,-3,fb,-1,qcd*3,fd)
     &		     *Propvp(fa,fb,3)*Propvp(fc,fd,3)

	if (qef.eq.0) then
	  do j=-1,1,2
	    Ce(j) =  a1*Propvp(fe,ff,1)*Cvf(fs,-j,1,fe)*Cvvv(1,-3,3)
     &	            +z1*Propvp(fe,ff,2)*Cvf(fs,-j,2,fe)*Cvvv(2,-3,3)
      	    Cg(j) =  a2*Propvp(fg,fh,1)*Cvf(fs,-j,1,fg)*Cvvv(1,3,-3)
     &	            +z2*Propvp(fg,fh,2)*Cvf(fs,-j,2,fg)*Cvvv(2,3,-3)
	  enddo
	else
	  fac = fac*Propvp(fe,ff,3)*Propvp(fg,fh,3)
     &	  	   *Cwwvv(3,3,-3,-3)*Cvfvf(fs,-1,3,fe,-1,3,fg)
	  Ce(+1) = 1d0
      	  Cg(+1) = 1d0
	  Ce(-1) = 0d0
      	  Cg(-1) = 0d0
	endif

	do 200 i5 = -1,1,2
	do 200 i7 = -1,1,2
          A(i5,i7) = App(yef(i5),yef(-i5),ygh(i7),ygh(-i7))
200	continue

	 i(1) = +1
	 i(2) = -1
	 i(3) = +1
	 i(4) = -1
	do 100 i5 = -1,1,2
	 i(5) =  i5
	 i(6) = -i5
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) + fac*Ce(i5)*Cg(i7)*A(i5,i7)
100	continue

	n4a=n4a+1

	end

************************************************************************
        subroutine addM4b(M,fs,a2,z2,a4,z4)
************************************************************************
*                                   \       /
*                                fh _\|   _/ fg
*                                     \   /|
*                                      \ / 
*               \                       *             /
*                \_                     <           |/_
*             fa |\                     < V4        /  ff
*                  \    V1              <          / 
*                   *vvvvvvvvv*- - - - -*vvvvvvvvv*  
*                  /          <    S1        V3    \
*             fb _/        V2 <                    _\| fe
*                /|           <                      \
*               /             *                       \
*                            / \
*                       fc |/_  \_ fd
*                          /    |\ 
*                         /       \
*	
*	Fields: fermions fa...fh incoming 
*	        vector bosons flowing out of ffV vertex 
*
*	Insertion:  V1 V2 V3 V4 S1 = W- W+  W+ W-  H    with fa<fg
*	                             W- W+  Z  Z   H    with fe<fg
*	                             Z  Z   Z  Z   H    with fa<fc,fa<fe,
*							     fa<fg,fe<fg
*	                             W- A/Z W+ A/Z phi  
*	
*-----------------------------------------------------------------------
*       18.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),A(-1:1,-1:1),B(-1:1,-1:1),App,fac
	complex*16 Cc(-1:1),Cg(-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer v1,v3,xa,xb,xc,xd,i(8),j,k,i1,i3,i5,i7,a2,z2,a4,z4,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

 	App(xa,xb,xc,xd) = csp(xb,xd)*sp(xa,xc)

	if((.not.((qab.eq.-1).and.(qcd.eq.+1).and.(qef.eq. 1).and.
     &	          (fa.lt.fg))).and.
     &	   (.not.((qab.eq.-1).and.(qcd.eq.+1).and.(qef.eq. 0).and.
     &	          (fe.lt.fg))).and.
     &	   (.not.((qab.eq. 0).and.(qcd.eq. 0).and.(qef.eq. 0).and.
     &	          (fa.lt.fc).and.(fa.lt.fe).and.
     &		  (fa.lt.fg).and.(fe.lt.fg))).and.
     &	   (.not.((qab.eq.-1).and.(qcd.eq. 0).and.(qef.eq. 1)))) return

	do 200 j = -1,1,2
	do 200 k = -1,1,2
	  A(+j,+k) = App(yab(j),yab(-j),ycd(k),ycd(-k)) 
          B(+j,+k) = App(yef(j),yef(-j),ygh(k),ygh(-k))
200	continue

	if (qab+qcd.eq.0) then
	  v1 =  2-qab
	  v3 = -2-qef
	  fac = -sgn*4d0*Cvvs(v1,-v1,0)*Cvvs(v3,-v3,0)
     &		    *Propvp(fa,fb,v1)*Propvp(fc,fd,v1)
     &	            *Propvp(fe,ff,v3)*Propvp(fg,fh,v3)
     &		    *Prop(vp(fa,fb)+vp(fa,fc)+vp(fa,fd)+
     &	  	  	  vp(fb,fc)+vp(fb,fd)+vp(fc,fd),0)
	  do j=-1,1,2
	    Cc(j) = Cvf(fs,-j,v1,fc)
	    Cg(j) = Cvf(fs,-j,v3,fg)
	  enddo
	else
	  v1 = +3
	  v3 = -3
	  fac = -sgn*4d0*Propvp(fa,fb,3)*Propvp(fe,ff,3)
     &		    *Prop(vp(fa,fb)+vp(fa,fc)+vp(fa,fd)+
     &			  vp(fb,fc)+vp(fb,fd)+vp(fc,fd),3)
	  do j=-1,1,2
	    Cc(j) = a2*Cvf(fs,-j,1,fc)*Cvvs(+3,1,-3)*Propvp(fc,fd,1)
     &	           +z2*Cvf(fs,-j,2,fc)*Cvvs(+3,2,-3)*Propvp(fc,fd,2)
	    Cg(j) = a4*Cvf(fs,-j,1,fg)*Cvvs(-3,1,+3)*Propvp(fg,fh,1)
     &	           +z4*Cvf(fs,-j,2,fg)*Cvvs(-3,2,+3)*Propvp(fg,fh,2)
	  enddo
	endif

	do 100 i1 = -1,1,2
	 i(1) =  i1
	 i(2) = -i1
	do 100 i3 = -1,1,2
	 i(3) =  i3
	 i(4) = -i3
	do 100 i5 = -1,1,2
	 i(5) =  i5
	 i(6) = -i5
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) +fac*Cvfvf(fs,-i1,-v1,fa,-i5,-v3,fe)
     &	                *Cc(i3)*A(i1,i3)*Cg(i7)*B(i5,i7)
100	continue

	n4b=n4b+1

	end

************************************************************************
        subroutine addM4c(M,fs,a1,z1,a2,z2,a3,z3)
************************************************************************
*                                   \       /
*                                fh _\|   _/ fg
*                                     \   /|
*                                      \ / 
*               \                       *             /
*                \_                     <           |/_
*             fa |\                     < V2        /  ff
*                  \    W-              <          / 
*                   *vvv > vvv*vvvvvvvvv*vvv < vvv*  
*                  /          <    V3        W+    \
*             fb _/        V1 <                    _\| fe
*                /|           <                      \
*               /             *                       \
*                            / \
*                       fc |/_  \_ fd
*                          /    |\ 
*                         /       \
*	
*	Fields: fermions fa...fh incoming 
*	        boson V1,V2 flowing out of ffV vertex 
*	        boson V3 flowing from left to right
*	
*	Insertions:  V1 V2 V3 = W+  W-  A/Z  with fa<fg
*	                        A/Z A/Z W-
*-----------------------------------------------------------------------
*       18.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),A(-1:1,-1:1),App,fac,Cc(-1:1),Cg(-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer xc,xd,xg,xh,i(8),j,i3,i7,a1,z1,a2,z2,a3,z3,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

 	App(xc,xd,xg,xh) = 
     &	csp(fb,xd)*csp(ff,xh)*sp(fa,xc)*sp(fe,xg)*kk
     &	+2d0*csp(fb,xd)*sp(fa,xc)
     &	*( (bsp(ff,fa,fe)+bsp(ff,fb,fe))*(bsp(xh,xc,xg)+bsp(xh,xd,xg))
     &	  -(bsp(ff,xc,fe)+bsp(ff,xd,fe))*(bsp(xh,fa,xg)+bsp(xh,fb,xg)) )
     &	+2d0*csp(ff,xh)*sp(fe,xg)
     &	*( (bsp(fb,fe,fa)+bsp(fb,ff,fa))*(bsp(xd,xg,xc)+bsp(xd,xh,xc))
     &	  -(bsp(fb,xg,fa)+bsp(fb,xh,fa))*(bsp(xd,fe,xc)+bsp(xd,ff,xc)) )
     &	+2d0*(bsp(fb,xc,fa)+bsp(fb,xd,fa))
     &	*( csp(xd,xh)*sp(xc,xg)*(bsp(ff,xg,fe)+bsp(ff,xh,fe))
     &	  -csp(xd,ff)*sp(xc,fe)*(bsp(xh,fe,xg)+bsp(xh,ff,xg)) )
     &	-2d0*(bsp(xd,fa,xc)+bsp(xd,fb,xc))
     &	*( csp(fb,xh)*sp(fa,xg)*(bsp(ff,xg,fe)+bsp(ff,xh,fe))
     &	  -csp(fb,ff)*sp(fa,fe)*(bsp(xh,fe,xg)+bsp(xh,ff,xg)) )

	if( (qab.ne.-1).or.(qef.ne.1).or.
     &	    (.not.(((qcd.eq.1).and.(fa.lt.fg)).or.((qcd.eq.0)))) )
     &	  return

	kk = ( vp(fa,fe)+vp(fa,ff)+vp(fb,fe)+vp(fb,ff)
     &	      -vp(fc,fe)-vp(fc,ff)-vp(fd,fe)-vp(fd,ff)
     &	      -vp(fa,fg)-vp(fa,fh)-vp(fb,fg)-vp(fb,fh)
     &	      +vp(fc,fg)+vp(fc,fh)+vp(fd,fg)+vp(fd,fh) )/2d0

	do 200 i3 = -1,1,2
	do 200 i7 = -1,1,2
          A(i3,i7) = App(ycd(i3),ycd(-i3),ygh(i7),ygh(-i7))
200	continue

	fac =  sgn*4d0*Cvfvf(fs,-1,-3,fb,-1,3,ff)
     &		  *Propvp(fa,fb,3)*Propvp(fe,ff,3)

	if (qcd.eq.1) then
	  fac = fac*Cvfvf(fs,-1,3,fd,-1,-3,fh)
     &		   *Propvp(fg,fh,3)*Propvp(fc,fd,3)
     &	           *( a3*Cvvv(3,-3,1)*Cvvv(-3,3,1)
     &		        *Prop(vp(fa,fb)+vp(fa,fc)+vp(fa,fd)+
     &			      vp(fb,fc)+vp(fb,fd)+vp(fc,fd),1)
     &	             +z3*Cvvv(3,-3,2)*Cvvv(-3,3,2)
     &		        *Prop(vp(fa,fb)+vp(fa,fc)+vp(fa,fd)+
     &			      vp(fb,fc)+vp(fb,fd)+vp(fc,fd),2) )
	  Cc(+1) = 1d0
	  Cc(-1) = 0d0
	  Cg(+1) = 1d0
	  Cg(-1) = 0d0
	else
	  fac = fac*Prop(vp(fa,fb)+vp(fa,fc)+vp(fa,fd)+
     &		 	 vp(fb,fc)+vp(fb,fd)+vp(fc,fd),3)
	  do j=-1,1,2
	  Cc(j) =  a1*Cvvv( 3,1,-3)*Cvf(fs,-j,1,fc)*Propvp(fc,fd,1)
     &	          +z1*Cvvv( 3,2,-3)*Cvf(fs,-j,2,fc)*Propvp(fc,fd,2)
	  Cg(j) =  a2*Cvvv(-3,1, 3)*Cvf(fs,-j,1,fg)*Propvp(fg,fh,1)
     &	          +z2*Cvvv(-3,2, 3)*Cvf(fs,-j,2,fg)*Propvp(fg,fh,2)
	  enddo
	endif

	 i(1) =  1
	 i(2) = -1
	 i(5) =  1
	 i(6) = -1
	do 100 i3 = -1,1,2
	 i(3) =  i3
	 i(4) = -i3
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) + fac*Cc(i3)*Cg(i7)*A(i3,i7)
100	continue

	n4c=n4c+1

	end

************************************************************************
        subroutine addM3a(M,fs,a1,z1,g1,a2,z2,g2,a3,z3,g3)
************************************************************************
*                           
*               \                 /
*                \_             _/
*             fa |\             /| fc
*                  \    V1     /
*                   *vvvvvvvvv*--<-- fd
*                   |
*                   |             /
*               f1 /|\          _/
*                   |           /| fe
*                   |   V2     /
*                   *vvvvvvvvv*
*                   |          \
*                   |           \_ ff
*               f2 /|\          |\ 
*                   |             \
*                   |   V3
*                   *vvvvvvvvv*-->-- fg
*                  /           \
*             fb _/             \_ fh
*                /|             |\
*               /                 \
*	
*	Fields: fermions fa...fh incoming 
*	        bosons V1V2V3 flowing to the left
*-----------------------------------------------------------------------
*       18.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 Cv2f2(-1:1)
	complex*16 M(0:255,24),A(-1:1,-1:1,-1:1,-1:1)
	complex*16 Am(-1:1,-1:1,-1:1,-1:1),Ap(-1:1,-1:1,-1:1,-1:1)
	complex*16 Amm,Amp,fac,Cac(-1:1,-1:1),Cbg(-1:1,-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop
	complex*16 m1m2,aux
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer v1,v2,v3,v4,v5,xc,xd,xe,xf,xg,xh,i(8),j,k,i1,i3,i5,i7
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1)
	integer a1,z1,g1,a2,z2,g2,a3,z3,g3,ccl,fs
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

	Amm(xc,xd,xe,xf,xg,xh) = csp(fb,xh)*sp(fa,xc)
     &	 *(-bsp(xd,fa,xe)-bsp(xd,xc,xe))*( bsp(xf,fb,xg)+bsp(xf,xh,xg))

	Amp(xc,xd,xe,xf,xg,xh) = csp(fb,xh)*sp(fa,xc)
     &	 *csp(xd,xf)*sp(xg,xe) 

	m1m2 = mf(fv(fs,fa,-qcd*3))*mf(fv(fs,fb,-qgh*3))

	if (qcd.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cac(j,k) =
     &		  (a1*Cvfvf(fs,-j,1,fa,-k,1,fc)+g1*gs2)*Propvp(fc,fd,1) 
     &	        + z1*Cvfvf(fs,-j,2,fa,-k,2,fc)*Propvp(fc,fd,2)
	  enddo
	  enddo
	else
	  if (g1.eq.1) return
	  Cac(+1,+1) =Cvfvf(fs,-1,-qcd*3,fa,-1,qcd*3,fc)*Propvp(fc,fd,3)
	  Cac(+1,-1) = 0d0
	  Cac(-1,+1) = 0d0
	  Cac(-1,-1) = 0d0
	endif

	if (qgh.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cbg(j,k) =
     &		  (a3*Cvfvf(fs,+j,1,fb,-k,1,fg)+g3*gs2)*Propvp(fg,fh,1)
     &	         +z3*Cvfvf(fs,+j,2,fb,-k,2,fg)*Propvp(fg,fh,2)
	  enddo
	  enddo
	else
	  if (g3.eq.1) return
	  Cbg(-1,+1) =Cvfvf(fs,-1,-qgh*3,fb,-1,qgh*3,fg)*Propvp(fg,fh,3)
	  Cbg(-1,-1) = 0d0
	  Cbg(+1,+1) = 0d0
	  Cbg(+1,-1) = 0d0
	endif

	do 200 i3=-1,1,2
	do 200 i5=-1,1,2
	do 200 i7=-1,1,2
	   aux = Amm(ycd(i3),ycd(-i3),yef(i5),yef(-i5),ygh(i7),ygh(-i7))
	   Am(+1, i3, i5, i7) = aux
	   Am(-1,-i3,-i5,-i7) = dconjg(aux)
	   Ap(-1,-i3,-i5,-i7) = m1m2*dconjg(aux)
200	continue

	if (m1m2.ne.dcmplx(0d0,0d0)) then
	do 210 i3=-1,1,2
	do 210 i5=-1,1,2
	do 210 i7=-1,1,2
	   aux = Amp(ycd(i3),ycd(-i3),yef(i5),yef(-i5),ygh(i7),ygh(-i7))
	   Ap(+1, i3, i5, i7) = m1m2*aux
	   Ap(-1,-i3,-i5,-i7) = m1m2*dconjg(aux)
210	continue
	endif

	fac=sgn*8d0*Prop(vp(fa,fc)+vp(fa,fd)+vp(fc,fd),fv(fs,fa,-qcd*3))
     &	           *Prop(vp(fb,fg)+vp(fb,fh)+vp(fg,fh),fv(fs,fb,-qgh*3))

	do 100 i1 = -1,1,2
	 i(1) =  i1
	 i(2) = -i1
	do 100 i3 = -1,1,2
	 i(3) =  i3
	 i(4) = -i3
	do 100 i5 = -1,1,2
	 i(5) =  i5
	 i(6) = -i5
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 if ((qef.eq.0).and.(qcd.eq.0)) then
          A(i1,i3,i5,i7) = Am(i1,i3,i5,i7)*(
     &	    (a2*Cvfvf(fs,-i1,1,fa,-i5,1,fe)+g2*gs2)*Propvp(fe,ff,1)
     &	    +z2*Cvfvf(fs,-i1,2,fa,-i5,2,fe)*Propvp(fe,ff,2) )
	 elseif ((qef.eq.0).and.(qcd.ne.0)) then
          A(i1,i3,i5,i7) = Am(i1,i3,i5,i7)*(
     &	    (a2*Cvf(fs,-i5,1,fe)*Cvif(fs,-i1,1,fa)+g2*gs2)
     &		*Propvp(fe,ff,1) 
     &	    +z2*Cvf(fs,-i5,2,fe)*Cvif(fs,-i1,2,fa)*Propvp(fe,ff,2) )
	  if (m1m2.ne.dcmplx(0d0,0d0))
     &    A(i1,i3,i5,i7) = A(i1,i3,i5,i7) + Ap(i1,i3,i5,i7)*(
     &	    (a2*Cvf(fs,-i5,1,fe)*Cvif(fs,+i1,1,fa)+g2*gs2)
     &		*Propvp(fe,ff,1)
     &	    +z2*Cvf(fs,-i5,2,fe)*Cvif(fs,+i1,2,fa)*Propvp(fe,ff,2) )
	 elseif ((qef.ne.0).and.(qcd.eq.0)) then
	  if (g2.eq.1) return
          A(i1,i3,i5,i7) = Cvfvf(fs,-i1,-qef*3,fa,-i5,qef*3,fe)
     &			     *Propvp(fe,ff,3)*Am(i1,i3,i5,i7)
	 else
	  if (g2.eq.1) return
          A(i1,i3,i5,i7) =  Cvf(fs,-i5, qef*3,fe)*Propvp(fe,ff,3)
     &                    *Cvif(fs,-i1,-qef*3,fa)*Am(i1,i3,i5,i7)
	 endif
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) + fac*Cac(i1,i3)*Cbg(-i1,i7)
     &					  *A(i1,i3,i5,i7)
100	continue

	n3a=n3a+1

	end

************************************************************************
        subroutine addM3b(M,fs,a1,z1)
************************************************************************
*                           
*               \                 /
*                \_             _/
*             fa |\             /| fc
*                  \    W-     /           
*                   *vv < vvvv*--<-- fd   
*                   |       
*                   |                       /
*              top /|\                    _/ 
*                   |                     /| fe
*                   |  phi-       W+     /
*                   *- ->- - -*vvv < vvv*--<-- ff
*                  /           |_
*             fb _/              |_
*                /|            V1  |_
*               /                    *-->-- fg
*	                              \
*	                               \_ 
*	                               |\ fh
*	                                 \
*	
*	Fields: fermions fa...fh incoming 
*-----------------------------------------------------------------------
*       18.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),B(-1:1),Bp,fac,Cg(-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer xg,xh,i(8),i7,a1,z1,j,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

 	Bp(xg,xh) =  csp(ff,xh)*sp(fe,xg)

	if ((fv(fs,fa,3).ne.4).or.(qab.ne.0).or.(qcd.ne.-1).or.
     &	    (qef.ne.1)) return

	do 200 i7 = -1,1,2
	  B(i7)  = Bp(ygh(i7),ygh(-i7)) 
	  Cg(i7) =  a1*Cvvs(1,-3,3)*Cvf(fs,-i7,1,fg)*Propvp(fg,fh,1)
     &	           +z1*Cvvs(2,-3,3)*Cvf(fs,-i7,2,fg)*Propvp(fg,fh,2)
200	continue

	fac =  sgn*4d0*Csbt*Cvfvf(fs,-1,3,fa,-1,-3,fc)*Cvf(fs,-1,3,fe)
     &		      *Propvp(fc,fd,3)*Propvp(fe,ff,3)
     &	              *Prop( vp(fa,fb)+vp(fa,fc)+vp(fa,fd)
     &	                    +vp(fb,fc)+vp(fb,fd)+vp(fc,fd),3)
     &	              *Prop(vp(fa,fc)+vp(fa,fd)+vp(fc,fd),4)
     &		      *csp(fb,fd)*sp(fa,fc)*mf(4)

	 i(1) = +1
	 i(2) = -1 
	 i(3) = +1
	 i(4) = -1
	 i(5) = +1
	 i(6) = -1
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) + fac*B(i7)*Cg(i7)
100	continue

	n3b=n3b+1

	end

************************************************************************
        subroutine addM3c(M,fs,a1,z1)
************************************************************************
*                           
*                                        /
*                                      |/_ fh
*                                      /
*                                     /
*               \                   _*-->-- fg
*                \_            V1 _|
*             fa |\             _|
*                  \   phi+    |           
*                   *- ->- - -*vvv < vvv*--<-- ff
*                   |             W-     \
*                   |                    _\| fe
*              top /|\                     \
*                   |                       \
*                   |   W+       
*                   *vv < vvvv*--<-- fd
*                  /           \
*             fb _/            _\| fc
*                /|              \
*               /                 \
*	
*	Fields: fermions fa...fh incoming 
*-----------------------------------------------------------------------
*       18.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),B(-1:1),Bp,fac,Cg(-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer xe,xf,xg,xh,i(8),j,i1,i3,i5,i7,a1,z1,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

 	Bp(xg,xh) =  csp(ff,xh)*sp(fe,xg)

	if ((fv(fs,fb,-3).ne.4).or.(qab.ne.0).or.(qcd.ne.1).or.
     &	    (qef.ne.-1)) return

	do 200 i7 = -1,1,2
	  B(i7)  = Bp(ygh(i7),ygh(-i7)) 
	  Cg(i7) =  a1*Cvvs(1,3,-3)*Cvf(fs,-i7,1,fg)*Propvp(fg,fh,1)
     &	           +z1*Cvvs(2,3,-3)*Cvf(fs,-i7,2,fg)*Propvp(fg,fh,2)
200	continue

	fac =  sgn*4d0*Csbt*Cvfvf(fs,-1,-3,fb,-1,3,fc)*Cvf(fs,-1,-3,fe)
     &		      *Propvp(fc,fd,3)*Propvp(fe,ff,3)
     &	              *Prop( vp(fa,fb)+vp(fa,fc)+vp(fa,fd)
     &	                    +vp(fb,fc)+vp(fb,fd)+vp(fc,fd),3)
     &	              *Prop(vp(fb,fc)+vp(fb,fd)+vp(fc,fd),4)
     &		      *csp(fb,fd)*sp(fa,fc)*mf(4)

	 i(1) = +1 
	 i(2) = -1 
	 i(3) = +1 
	 i(4) = -1 
	 i(5) = +1
	 i(6) = -1
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) + fac*B(i7)*Cg(i7)
100	continue

	n3c=n3c+1

	end

************************************************************************
        subroutine addM3d(M,fs,a1,z1,g1,a2,z2)
************************************************************************
*                           
*               \                 /
*                \_             _/
*             fa |\             /| fc
*                  \    V1     /           
*                   *vvvvvvvvv*--<-- fd   
*                   |       
*                   |                       /
*               f1 /|\                    _/ 
*                   |                     /| fe
*                   |   V2         W     /
*                   *vvvvvvvvv*vvvvvvvvv*--<-- ff
*                  /           |_
*             fb _/              |_
*                /|            V3  |_
*               /                    *-->-- fg
*	                              \
*	                               \_ 
*	                               |\ fh
*	                                 \
*	
*	Fields: fermions fa...fh incoming 
*	        bosons V1,V2,V3 flowing to the left
*	
*	Insertions: V1 V2 V3 W  = W+- W-+ A/Z W-+
*	                          W+- A/Z W+  W-
*	                          A/Z W-+ A/Z W-+
*	                          A/Z A/Z W+  W-
*	
*-----------------------------------------------------------------------
*       18.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),A(-1:1,-1:1,-1:1),Am,fac
	complex*16 Cac(-1:1,-1:1),Cbg(-1:1,-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop,aux
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer xc,xd,xe,xf,xg,xh,i(8),j,k,i1,i3,i5,i7,a1,z1,g1,a2,z2,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

	Am(xc,xd,xe,xf,xg,xh) = sp(xc,fa)
     &	  *( csp(xf,xh)*sp(xe,xg)*( 
     &	       csp(fa,xd)*( bsp(fb,xe,fa)+bsp(fb,xf,fa)
     &	                   -bsp(fb,xg,fa)-bsp(fb,xh,fa))
     &	      +csp(xc,xd)*( bsp(fb,xe,xc)+bsp(fb,xf,xc)
     &	                   -bsp(fb,xg,xc)-bsp(fb,xh,xc)) )
     &	    +2d0*csp(fb,xh)*(bsp(xd,fa,xg)+bsp(xd,xc,xg))
     &	                   *(bsp(xf,xg,xe)+bsp(xf,xh,xe))
     &	    -2d0*csp(fb,xf)*(bsp(xd,fa,xe)+bsp(xd,xc,xe))
     &	                   *(bsp(xh,xe,xg)+bsp(xh,xf,xg))  )

	if ((qef.eq.0).or.((qef.eq.1).and.(qgh.eq.-1))) return

	do 200 i3 = -1,1,2
	do 200 i7 = -1,1,2
	  A(+1, i3, i7) =
     &	       Am(ycd(i3),ycd(-i3),yef(+1),yef(-1),ygh(i7),ygh(-i7))
      	  A(-1,-i3,-i7) = dconjg( 
     &	       Am(ycd(i3),ycd(-i3),yef(-1),yef(+1),ygh(i7),ygh(-i7)) )
200	continue

	fac = -sgn*4d0*Cvf(fs,-1,qef*3,fe)*Propvp(fe,ff,3)
     &	        *Prop(vp(fa,fc)+vp(fa,fd)+vp(fc,fd),fv(fs,fa,-qcd*3))

	if (qcd.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cac(j,k) =
     &		  (a1*Cvfvf(fs,-j,1,fa,-k,1,fc)+g1*gs2)*Propvp(fc,fd,1)
     &	         +z1*Cvfvf(fs,-j,2,fa,-k,2,fc)*Propvp(fc,fd,2)
	  enddo
	  enddo
	else
	  if (g1.eq.1) return
	  Cac(+1,+1) =Cvfvf(fs,-1,-qcd*3,fa,-1,qcd*3,fc)*Propvp(fc,fd,3)
	  Cac(+1,-1) = 0d0
	  Cac(-1,+1) = 0d0
	  Cac(-1,-1) = 0d0
	endif

	if (qgh.eq.0) then
	  fac = fac*qef*Cvf(fs,-1,-qef*3,fb)
     &	           *Prop( vp(fa,fb)+vp(fa,fc)+vp(fa,fd)
     &	                 +vp(fb,fc)+vp(fb,fd)+vp(fc,fd),3)
	  do k=-1,1,2
	    Cbg(+1,k) = 0d0
	    Cbg(-1,k) = a2*Cvvv(3,-3,1)*Cvf(fs,-k,1,fg)*Propvp(fg,fh,1)
     &	               +z2*Cvvv(3,-3,2)*Cvf(fs,-k,2,fg)*Propvp(fg,fh,2)
	  enddo
	else
	  fac = fac*Cvf(fs,-1,3,fg)*Propvp(fg,fh,3)
	  do j=-1,1,2
	    Cbg(j,-1) = 0d0
	    Cbg(j,+1) = a2*Cvvv(1,3,-3)*Cvf(fs,j,1,fb)
     &	                  *Prop( vp(fa,fb)+vp(fa,fc)+vp(fa,fd)
     &	                        +vp(fb,fc)+vp(fb,fd)+vp(fc,fd),1)
     &	               +z2*Cvvv(2,3,-3)*Cvf(fs,j,2,fb)
     &	                  *Prop( vp(fa,fb)+vp(fa,fc)+vp(fa,fd)
     &	                        +vp(fb,fc)+vp(fb,fd)+vp(fc,fd),2)
	  enddo
	endif

	 i(5) = +1 
	 i(6) = -1
	do 100 i1 = -1,1,2
	 i(1) =  i1
	 i(2) = -i1
	do 100 i3 = -1,1,2
	 i(3) =  i3
	 i(4) = -i3
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) +fac*Cac(i1,i3)*Cbg(-i1,i7)*A(i1,i3,i7)
100	continue

	n3d=n3d+1

	end

************************************************************************
        subroutine addM3e(M,fs,a1,z1,g1,a2,z2)
************************************************************************
*                           
*                                        /
*                                      |/_ fh
*                                      /
*                                     /
*               \                   _*-->-- fg
*                \_            V3 _|
*             fa |\             _|
*                  \    V2     |           
*                   *vvvvvvvvv*vvvvvvvvv*--<-- ff
*                   |              W     \
*                   |                    _\| fe
*               f1 /|\                     \
*                   |                       \
*                   |   V1       
*                   *vvvvvvvvv*--<-- fd
*                  /           \
*             fb _/            _\| fc
*                /|              \
*               /                 \
*	
*	Fields: fermions fa...fh incoming 
*	        bosons V1,V2,V3 flowing to the left
*	
*	Insertions: V1 V2 V3 W  = W+- W-+ A/Z W-+
*	                          W+- A/Z W+  W-
*	                          A/Z W-+ A/Z W-+
*	                          A/Z A/Z W+  W-
*	
*-----------------------------------------------------------------------
*       17.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),A(-1:1,-1:1,-1:1),Am,fac
	complex*16 Cbc(-1:1,-1:1),Cag(-1:1,-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop,aux
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer xc,xd,xe,xf,xg,xh,i(8),j,k,i1,i3,i5,i7,a1,z1,g1,a2,z2,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

	Am(xc,xd,xe,xf,xg,xh) = csp(xd,fb)
     &	  *( csp(xf,xh)*sp(xe,xg)*( 
     &	       sp(xc,fb)*( bsp(fb,xe,fa)+bsp(fb,xf,fa)
     &	                  -bsp(fb,xg,fa)-bsp(fb,xh,fa))
     &	      +sp(xc,xd)*( bsp(xd,xe,fa)+bsp(xd,xf,fa)
     &	                  -bsp(xd,xg,fa)-bsp(xd,xh,fa)) )
     &	    +2d0*sp(xg,fa)*(bsp(xh,fb,xc)+bsp(xh,xd,xc))
     &	                  *(bsp(xf,xg,xe)+bsp(xf,xh,xe))
     &	    -2d0*sp(xe,fa)*(bsp(xf,fb,xc)+bsp(xf,xd,xc))
     &	                  *(bsp(xh,xe,xg)+bsp(xh,xf,xg))  )

	if ((qef.eq.0).or.((qef.eq.1).and.(qgh.eq.-1))) return

	do 200 i3 = -1,1,2
	do 200 i7 = -1,1,2
	  A(+1, i3, i7) =
     &	       Am(ycd(i3),ycd(-i3),yef(+1),yef(-1),ygh(i7),ygh(-i7))
      	  A(-1,-i3,-i7) = dconjg( 
     &	       Am(ycd(i3),ycd(-i3),yef(-1),yef(+1),ygh(i7),ygh(-i7)) )
200	continue

	fac = -sgn*4d0*Cvf(fs,-1,qef*3,fe)*Propvp(fe,ff,3)
     &	          *Prop(vp(fb,fc)+vp(fb,fd)+vp(fc,fd),fv(fs,fb,-qcd*3))

	if (qcd.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cbc(j,k) =
     &		  (a1*Cvfvf(fs,j,1,fb,-k,1,fc)+g1*gs2)*Propvp(fc,fd,1)
     &	         +z1*Cvfvf(fs,j,2,fb,-k,2,fc)*Propvp(fc,fd,2)
	  enddo
	  enddo
	else
	  if (g1.eq.1) return
	  Cbc(-1,+1) =Cvfvf(fs,-1,-qcd*3,fb,-1,qcd*3,fc)*Propvp(fc,fd,3)
	  Cbc(-1,-1) = 0d0
	  Cbc(+1,+1) = 0d0
	  Cbc(+1,-1) = 0d0
	endif

	if (qgh.eq.0) then
	  fac = fac*qef*Cvf(fs,-1,-qef*3,fa)
     &	           *Prop( vp(fa,fb)+vp(fa,fc)+vp(fa,fd)
     &	                 +vp(fb,fc)+vp(fb,fd)+vp(fc,fd),3)
	  do k=-1,1,2
	    Cag(-1,k) = 0d0
	    Cag(+1,k) = a2*Cvvv(3,-3,1)*Cvf(fs,-k,1,fg)*Propvp(fg,fh,1)
     &	               +z2*Cvvv(3,-3,2)*Cvf(fs,-k,2,fg)*Propvp(fg,fh,2)
	  enddo
	else
	  fac = fac*Cvf(fs,-1,3,fg)*Propvp(fg,fh,3)
	  do j=-1,1,2
	    Cag(j,-1) = 0d0
	    Cag(j,+1) = a2*Cvvv(1,3,-3)*Cvf(fs,-j,1,fa)
     &	                  *Prop( vp(fa,fb)+vp(fa,fc)+vp(fa,fd)
     &	                        +vp(fb,fc)+vp(fb,fd)+vp(fc,fd),1)
     &	               +z2*Cvvv(2,3,-3)*Cvf(fs,-j,2,fa)
     &	                  *Prop( vp(fa,fb)+vp(fa,fc)+vp(fa,fd)
     &	                        +vp(fb,fc)+vp(fb,fd)+vp(fc,fd),2)
	  enddo
	endif

	 i(5) = +1 
	 i(6) = -1
	do 100 i1 = -1,1,2
	 i(1) =  i1
	 i(2) = -i1
	do 100 i3 = -1,1,2
	 i(3) =  i3
	 i(4) = -i3
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) +fac*Cbc(-i1,i3)*Cag(i1,i7)*A(i1,i3,i7)
100	continue

	n3e=n3e+1

	end

************************************************************************
        subroutine addM2a(M,fs)
************************************************************************
*                           
*               \                 /
*                \_             _/
*             fa |\             /| fe
*                  \    W-     /
*                   *vv < vvvv*--<-- ff
*                   |
*                   |  
*              top /|\ 
*                   |  
*                   |  phi-
*           fb -->--*- ->- - -*-->-- fc
*                             |
*                             |  
*                        top /|\ 
*                             |  
*                             |   
*           fg --<--*vv > vvvv*
*                  /    W+     \
*                _/             \_ fd
*             fh /|             |\ 
*               /                 \
*	
*	Fields: fermions fa...fh incoming 
*-----------------------------------------------------------------------
*       17.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),A(-1:1,-1:1,-1:1,-1:1,-1:1),Am,Ap,fac
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer i(8),j,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1),ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

	if ((qab.ne.0).or.(qef.ne.-1).or.(qcd.ne.0).or.
     &	    (fv(fs,fa,3).ne.4).or.(fv(fs,fc,3).ne.4)) return

	fac = -sgn*4d0*Propvp(fe,ff,3)*Propvp(fg,fh,3)
     &	              *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                    +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),3)
     &	              *Prop(vp(fa,fe)+vp(fa,ff)+vp(fe,ff),4)
     &	              *Prop(vp(fd,fg)+vp(fd,fh)+vp(fg,fh),4)
     &	       *Csbt**2*mf(fv(fs,fa,3))**2
     &	       *Cvfvf(fs,-1,3,fa,-1,-3,fe)*Cvfvf(fs,-1,-3,fd,-1,3,fg)
     &	       *csp(fb,ff)*sp(fa,fe)*csp(fd,fh)*sp(fc,fg)

	i(1) = +1
	i(2) = -1
	i(3) = +1
	i(4) = -1
	i(5) = +1
	i(6) = -1
	i(7) = +1
	i(8) = -1
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) + fac

	n2a=n2a+1

	end

************************************************************************
        subroutine addM2b(M,fs,a1,z1,g1,a2,z2,g2,a3,z3,g3)
************************************************************************
*                           
*               \                 /
*                \_             _/
*             fa |\             /| fe
*                  \    V1     /
*                   *vvvvvvvvv*--<-- ff
*                   |
*                   |  
*               f1 /|\ 
*                   |  
*                   |   V2 
*           fb -->--*vvvvvvvvv*-->-- fc
*                             |
*                             |  
*                         f2 /|\ 
*                             |  
*                             |   
*           fg --<--*vvvvvvvvv*
*                  /    V3     \
*                _/             \_ fd
*             fh /|             |\ 
*               /                 \
*	
*	Fields: fermions fa...fh incoming 
*-----------------------------------------------------------------------
*       19.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),A(-1:1,-1:1,-1:1,-1:1),Apmm,Ampm,fac
	complex*16 Cae(-1:1,-1:1),Cdg(-1:1,-1:1),Cbc(-1:1,-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop,aux
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer i(8),j,k,v1,v2,v3,i1,i3,i5,i7,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
        integer xe,xf,xg,xh,a1,z1,g1,a2,z2,g2,a3,z3,g3,ccl
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1)
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

	Apmm(xe,xf,xg,xh) = sp(xe,fa)*csp(fd,xh)
     &	  *(-bsp(xf,fa,fc)-bsp(xf,xe,fc))*(bsp(fb,fd,xg)+bsp(fb,xh,xg))
	Ampm(xe,xf,xg,xh) = csp(xf,fa)*sp(fb,fc)*csp(xh,fd)
     &	  *( sp(fa,xe)*(bsp(fa,fd,xg)+bsp(fa,xh,xg))
     &	    +sp(xf,xe)*(bsp(xf,fd,xg)+bsp(xf,xh,xg)) )

	if ((abs(qab+qef).gt.1).or.(abs(qef-q(fs,fa)).gt.1d0).or.
     &	    (abs(qgh+q(fs,fd)).gt.1d0)) return

	do 300 i5=-1,1,2
	do 300 i7=-1,1,2
	  A(+1,+1,i5,i7) = Apmm(yef(i5),yef(-i5),ygh(i7),ygh(-i7))
	  A(-1,+1,i5,i7) = Ampm(yef(i5),yef(-i5),ygh(i7),ygh(-i7))
	  do j=-1,1,2
	  A(-j,-1,-i5,-i7) = dconjg( A(+j,+1,i5,i7) )
	  enddo
300	continue

	fac=sgn*8d0*Prop(vp(fa,fe)+vp(fa,ff)+vp(fe,ff),fv(fs,fa,-qef*3))
     &	           *Prop(vp(fd,fg)+vp(fd,fh)+vp(fg,fh),fv(fs,fd,-qgh*3))

	if (qef.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cae(j,k) =
     &		  (a1*Cvfvf(fs,-j,1,fa,-k,1,fe)+g1*gs2)*Propvp(fe,ff,1)
     &	         +z1*Cvfvf(fs,-j,2,fa,-k,2,fe)*Propvp(fe,ff,2)
	  enddo
	  enddo
	else
	  if (g1.eq.1) return
	  Cae(+1,+1) =Cvfvf(fs,-1,-qef*3,fa,-1,qef*3,fe)*Propvp(fe,ff,3)
	  Cae(+1,-1) = 0d0
	  Cae(-1,+1) = 0d0
	  Cae(-1,-1) = 0d0
	endif

	if (qab+qef.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cbc(j,k) = (a2*Cvfvf(fs,+j,1,fb,-k,1,fc)+g2*gs2)
     &	                  *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                        +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),1)
     &	               +z2*Cvfvf(fs,+j,2,fb,-k,2,fc)
     &	                  *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                        +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),2)
	  enddo
	  enddo
	else
	  if (g2.eq.1) return
	  Cbc(-1,+1) = Cvfvf(fs,-1,(qab+qef)*3,fb,-1,-(qab+qef)*3,fc)
     &	                 *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                       +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),3)
	  Cbc(-1,-1) = 0d0
	  Cbc(+1,+1) = 0d0
	  Cbc(+1,-1) = 0d0
	endif

	if (qgh.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cdg(j,k) =
     &		  (a3*Cvfvf(fs,+j,1,fd,-k,1,fg)+g3*gs2)*Propvp(fg,fh,1)
     &	         +z3*Cvfvf(fs,+j,2,fd,-k,2,fg)*Propvp(fg,fh,2)
	  enddo
	  enddo
	else
	  if (g3.eq.1) return
	  Cdg(-1,+1) =Cvfvf(fs,-1,-qgh*3,fd,-1,qgh*3,fg)*Propvp(fg,fh,3)
	  Cdg(-1,-1) = 0d0
	  Cdg(+1,+1) = 0d0
	  Cdg(+1,-1) = 0d0
	endif

	do 100 i1 = -1,1,2
	 i(1) =  i1
	 i(2) = -i1
	do 100 i3 = -1,1,2
	 i(3) =  i3
	 i(4) = -i3
	do 100 i5 = -1,1,2
	 i(5) =  i5
	 i(6) = -i5
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) 
     &	  + fac*Cae(i1,i5)*Cbc(-i1,i3)*Cdg(-i3,i7)*A(i1,i3,i5,i7)
100	continue

	n2b=n2b+1

	end

************************************************************************
        subroutine addM2c(M,fs,a1,z1,g1,a2,z2,g2,a3,z3,g3)
************************************************************************
*                           
*               \                 /
*                \_             _/
*             fa |\             /| fe
*                  \    V1     /
*                   *vvvvvvvvv*--<-- ff
*                   |
*                   |  
*               f1 /|\ 
*                   |  
*                   |   V2  
*           fb -->--*vvvvvvvvv*--<-- fd
*                             |
*                             |  
*                         f2 \|/ 
*                             |  
*                             |   
*           fg --<--*vvvvvvvvv*
*                  /    V3     \
*                _/            _\| fc
*             fh /|              \ 
*               /                 \
*	
*	Fields: fermions fa...fh incoming     with fa<fc
*-----------------------------------------------------------------------
*       19.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),A(-1:1,-1:1,-1:1,-1:1),Apmm,Ampm,fac
	complex*16 Cae(-1:1,-1:1),Ccg(-1:1,-1:1),Cbd(-1:1,-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop,aux
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer i(8),j,k,v1,v2,v3,i1,i3,i5,i7,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1)
        integer xe,xf,xg,xh,a1,z1,g1,a2,z2,g2,a3,z3,g3,ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

	Apmm(xe,xf,xg,xh) = sp(xe,fa)*csp(fb,fd)*sp(xg,fc)
     &	  *( csp(fa,xf)*(-bsp(xh,fc,fa)-bsp(xh,xg,fa))
     &	    +csp(xe,xf)*(-bsp(xh,fc,xe)-bsp(xh,xg,xe)) )
	Ampm(xe,xf,xg,xh) = csp(xf,fa)*sp(fc,xg)
     &	  *(-bsp(fd,fa,xe)-bsp(fd,xf,xe))*(-bsp(xh,fc,fb)-bsp(xh,xg,fb))

	if ((abs(qab+qef).gt.1).or.(abs(qef-q(fs,fa)).gt.1d0).or.
     &	    (abs(qgh-q(fs,fc)).gt.1d0).or.(fa.gt.fc)) return

	do 300 i5=-1,1,2
	do 300 i7=-1,1,2
	  A(+1,+1,i5,i7) = Apmm(yef(i5),yef(-i5),ygh(i7),ygh(-i7))
	  A(-1,+1,i5,i7) = Ampm(yef(i5),yef(-i5),ygh(i7),ygh(-i7))
	  do j=-1,1,2
	  A(-j,-1,-i5,-i7) = dconjg( A(+j,+1,i5,i7) )
	  enddo
300	continue

	fac=sgn*8d0*Prop(vp(fa,fe)+vp(fa,ff)+vp(fe,ff),fv(fs,fa,-qef*3))
     &	           *Prop(vp(fc,fg)+vp(fc,fh)+vp(fg,fh),fv(fs,fc,-qgh*3))

	if (qef.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cae(j,k) =
     &		  (a1*Cvfvf(fs,-j,1,fa,-k,1,fe)+g1*gs2)*Propvp(fe,ff,1)
     &	         +z1*Cvfvf(fs,-j,2,fa,-k,2,fe)*Propvp(fe,ff,2)
	  enddo
	  enddo
	else
	  if (g1.eq.1) return
	  Cae(+1,+1) =Cvfvf(fs,-1,-qef*3,fa,-1,qef*3,fe)*Propvp(fe,ff,3)
	  Cae(+1,-1) = 0d0
	  Cae(-1,+1) = 0d0
	  Cae(-1,-1) = 0d0
	endif

	if (qab+qef.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cbd(j,k) = (a2*Cvfvf(fs,+j,1,fb,+k,1,fd)+g2*gs2)
     &	                  *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                        +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),1)
     &	               +z2*Cvfvf(fs,+j,2,fb,+k,2,fd)
     &	                  *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                        +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),2)
	  enddo
	  enddo
	else
	  if (g2.eq.1) return
	  Cbd(-1,-1) = Cvfvf(fs,-1,(qab+qef)*3,fb,-1,-(qab+qef)*3,fd)
     &	                 *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                       +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),3)
	  Cbd(-1,+1) = 0d0
	  Cbd(+1,-1) = 0d0
	  Cbd(+1,+1) = 0d0
	endif

	if (qgh.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Ccg(j,k) =
     &		  (a3*Cvfvf(fs,-j,1,fc,-k,1,fg)+g3*gs2)*Propvp(fg,fh,1)
     &	         +z3*Cvfvf(fs,-j,2,fc,-k,2,fg)*Propvp(fg,fh,2)
	  enddo
	  enddo
	else
	  if (g3.eq.1) return
	  Ccg(+1,+1) =Cvfvf(fs,-1,-qgh*3,fc,-1,qgh*3,fg)*Propvp(fg,fh,3)
	  Ccg(+1,-1) = 0d0
	  Ccg(-1,+1) = 0d0
	  Ccg(-1,-1) = 0d0
	endif

	do 100 i1 = -1,1,2
	 i(1) =  i1
	 i(2) = -i1
	do 100 i3 = -1,1,2
	 i(3) =  i3
	 i(4) = -i3
	do 100 i5 = -1,1,2
	 i(5) =  i5
	 i(6) = -i5
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) 
     &	  + fac*Cae(i1,i5)*Cbd(-i1,-i3)*Ccg(i3,i7)*A(i1,i3,i5,i7)
100	continue

	n2c=n2c+1

	end

************************************************************************
        subroutine addM2d(M,fs,a1,z1,g1,a2,z2,g2,a3,z3,g3)
************************************************************************
*                           
*               \                 /
*               _\|             _/
*             fb  \             /| fe
*                  \    V1     /
*                   *vvvvvvvvv*--<-- ff
*                   |
*                   |  
*               f1 \|/ 
*                   |  
*                   |   V2 
*           fa --<--*vvvvvvvvv*-->-- fc
*                             |
*                             |  
*                         f2 /|\ 
*                             |  
*                             |   
*           fg --<--*vvvvvvvvv*
*                  /    V3     \
*                _/             \_ fd
*             fh /|             |\ 
*               /                 \
*	
*	Fields: fermions fa...fh incoming    with fa<fc
*-----------------------------------------------------------------------
*       19.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	complex*16 M(0:255,24),A(-1:1,-1:1,-1:1,-1:1),Apmm,Ampm,fac
	complex*16 Cbe(-1:1,-1:1),Cdg(-1:1,-1:1),Cac(-1:1,-1:1)
	complex*16 sp(8,8),csp(8,8),bsp(8,8,8),propvp(8,8,-4:4),prop,aux
	complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
	complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
	complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
	complex*16 cvvs(-3:3,-3:3,-3:3)
	real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20),vp(8,8)
	integer i(8),j,k,v1,v2,v3,i1,i3,i5,i7,fs
	integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
	integer dir(8),pol(8),nfs
	integer fa,fb,fc,fd,fe,ff,fg,fh,qab,qcd,qef,qgh,sgn,npol
	integer n(8),yab(-1:1),ycd(-1:1),yef(-1:1),ygh(-1:1)
        integer xe,xf,xg,xh,a1,z1,g1,a2,z2,g2,a3,z3,g3,ccl
	integer n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d
	character*6 fname(20,8),dfname(20,8)

	common/perm/n,fa,fb,fc,fd,fe,ff,fg,fh,yab,ycd,yef,ygh,
     &	            qab,qcd,qef,qgh,sgn,ccl
	common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &		mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &		fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
	common/prods/sp,csp,bsp,vp,propvp
	common/count/n4a,n4b,n4c,n3a,n3b,n3c,n3d,n3e,n2a,n2b,n2c,n2d

	Apmm(xe,xf,xg,xh) = sp(fa,fc)*csp(xf,fb)*csp(xh,fd)
     &	  *( sp(xe,fb)*(bsp(fb,fd,xg)+bsp(fb,xh,xg))
     &	    +sp(xe,xf)*(bsp(xf,fd,xg)+bsp(xf,xh,xg)) )
	Ampm(xe,xf,xg,xh) = sp(fb,xe)*csp(xh,fd)
     &	  *(bsp(fa,fd,xg)+bsp(fa,xh,xg))*(bsp(xf,fb,fc)+bsp(xf,xe,fc))

	if ((abs(qab+qef).gt.1).or.(abs(qef+q(fs,fb)).gt.1d0).or.
     &	    (abs(qgh+q(fs,fd)).gt.1d0).or.(fa.gt.fc)) return

	do 300 i5=-1,1,2
	do 300 i7=-1,1,2
	  A(+1,+1,i5,i7) = Apmm(yef(i5),yef(-i5),ygh(i7),ygh(-i7))
	  A(-1,+1,i5,i7) = Ampm(yef(i5),yef(-i5),ygh(i7),ygh(-i7))
	  do j=-1,1,2
	  A(-j,-1,-i5,-i7) = dconjg( A(+j,+1,i5,i7) )
	  enddo
300	continue

	fac=sgn*8d0*Prop(vp(fb,fe)+vp(fb,ff)+vp(fe,ff),fv(fs,fb,-qef*3))
     &	           *Prop(vp(fd,fg)+vp(fd,fh)+vp(fg,fh),fv(fs,fd,-qgh*3))

	if (qef.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cbe(j,k) =
     &		  (a1*Cvfvf(fs,+j,1,fb,-k,1,fe)+g1*gs2)*Propvp(fe,ff,1)
     &	         +z1*Cvfvf(fs,+j,2,fb,-k,2,fe)*Propvp(fe,ff,2)
	  enddo
	  enddo
	else
	  if (g1.eq.1) return
	  Cbe(-1,+1) =Cvfvf(fs,-1,-qef*3,fb,-1,qef*3,fe)*Propvp(fe,ff,3)
	  Cbe(-1,-1) = 0d0
	  Cbe(+1,+1) = 0d0
	  Cbe(+1,-1) = 0d0
	endif

	if (qab+qef.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cac(j,k) = (a2*Cvfvf(fs,-j,1,fa,-k,1,fc)+g2*gs2)
     &	                  *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                        +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),1)
     &	               +z2*Cvfvf(fs,-j,2,fa,-k,2,fc)
     &	                  *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                        +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),2)
	  enddo
	  enddo
	else
	  if (g2.eq.1) return
	  Cac(+1,+1) = Cvfvf(fs,-1,(qab+qef)*3,fa,-1,-(qab+qef)*3,fc)
     &	                 *Prop( vp(fa,fb)+vp(fa,fe)+vp(fa,ff)
     &	                       +vp(fb,fe)+vp(fb,ff)+vp(fe,ff),3)
	  Cac(+1,-1) = 0d0
	  Cac(-1,+1) = 0d0
	  Cac(-1,-1) = 0d0
	endif

	if (qgh.eq.0) then
	  do j=-1,1,2
	  do k=-1,1,2
	    Cdg(j,k) =
     &		  (a3*Cvfvf(fs,+j,1,fd,-k,1,fg)+g3*gs2)*Propvp(fg,fh,1)
     &	         +z3*Cvfvf(fs,+j,2,fd,-k,2,fg)*Propvp(fg,fh,2)
	  enddo
	  enddo
	else
	  if (g3.eq.1) return
	  Cdg(-1,+1) =Cvfvf(fs,-1,-qgh*3,fd,-1,qgh*3,fg)*Propvp(fg,fh,3)
	  Cdg(-1,-1) = 0d0
	  Cdg(+1,+1) = 0d0
	  Cdg(+1,-1) = 0d0
	endif

	do 100 i1 = -1,1,2
	 i(1) =  i1
	 i(2) = -i1
	do 100 i3 = -1,1,2
	 i(3) =  i3
	 i(4) = -i3
	do 100 i5 = -1,1,2
	 i(5) =  i5
	 i(6) = -i5
	do 100 i7 = -1,1,2
	 i(7) =  i7
	 i(8) = -i7
	 npol=(255+i(n(1))*dir(1))/2
	 do j=2,8
	   npol = npol + i(n(j))*dir(j)*2**(j-2)
	 enddo
	 M(npol,ccl)=M(npol,ccl) 
     &	  + fac*Cbe(-i1,i5)*Cac(i1,i3)*Cdg(-i3,i7)*A(i1,i3,i5,i7)
100	continue

	n2d=n2d+1

	end

