************************************************************************
        program main
************************************************************************
*	Amplitude calculation for e+ e- --> 6 fermions
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	integer out,qalp

	common/out/out
	
C set output channel 
	out = 6

C set input data 
	qalp = 0
	call input(qalp)
	call paramout

C test output for e+e- --> 6f 
  	call testee6f

	end

************************************************************************
        subroutine input(qalp)
************************************************************************
*	input parameters of the Standard Model
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	integer qalp

        common/lparam/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,gh,me,me2,mt,mt2,gt

	pi      = 4d0*datan(1d0)

	alpha0  = 1d0/137.0359895d0
	alphaz  = 1d0/128.887d0
	alphas  = 0.119d0
	GF      = 1.16637D-5

	mw      = 80.35d0
	gw      = 2.05d0
	mz      = 91.1867d0
	gz      = 2.46d0
        mh      = 150d0
	gh      = 1d0
	mt      = 175d0
	gt      = 2d0
        me      = .51099907d-3  

c*** dependent parameters
	cw      = mw/mz
	cw2     = cw**2
	sw2     = 1d0-cw2
	sw      = sqrt(sw2)

        mw2     = mw**2
        mz2     = mz**2
        mh2     = mh**2
        mt2     = mt**2
        me2     = me**2

C*** input-parameter scheme
	if (qalp.eq.0) then
	  alpha = alpha0
	elseif (qalp.eq.1) then
	  alpha = alphaz
	elseif (qalp.eq.2) then
	  alpha = sqrt(2d0)*gf/pi*mw2*sw2
	endif

	el 	= sqrt(4d0*pi*alpha)

	end

************************************************************************
        subroutine paramout
************************************************************************
*	output of Standard-Model parameters
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
	integer out

        common/lparam/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,gh,me,me2,mt,mt2,gt
	common/out/out

1000	format(/,'Input parameters:',/,'-----------------',/,
     &	  2x,'alpha(0) = 1/', f10.6,',',
     &	  3x,'alpha(MZ) = 1/',f9.5,',',
     &	  3x,'alpha_s = ',f7.5,',',/,
     &	  8x,'GF = ',g12.7,',',/,
     &	  8x,'MW = ',f9.5,',',
     &	 13x,'MZ = ',f9.5,',',
     &	 10x,'MH = ',f9.5,',',/,
     &	  8x,'GW = ',f9.5,',',
     &	 13x,'GZ = ',f9.5,',',
     &	 10x,'GH = ',f9.5,',',/,
     &	  8x,'mt = ',f9.5,',',
     &	 13x,'me = ',g13.8,','/,
     &	  8x,'Gt = ',f9.5,'.')

	write(out,1000) 1d0/alpha0,1d0/alphaz,alphas,GF,
     &	  mw,mz,mh,gw,gz,gh,mt,me,gt

	end

************************************************************************
        subroutine procinit
************************************************************************
*	initialize assignment of final state
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
        complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
        complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
        complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
        complex*16 cvvs(-3:3,-3:3,-3:3)
        real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20)
        integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
        integer dir(8),pol(8),nfs
        integer i,fs
        character*6 gfname(0:1,-1:1,3),fname(20,8)
	character*6 dgfname(-1:1),dfname(20,8)

        common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &          mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &          fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname

C*** define particle names
        gfname(0,-1,1) = 'e'
        gfname(0,-1,2) = 'mu'
        gfname(0,-1,3) = 'tau'
        gfname(0,+1,1) = 'nu_e'
        gfname(0,+1,2) = 'nu_mu'
        gfname(0,+1,3) = 'nu_tau'
        gfname(1,-1,1) = 'd'
        gfname(1,-1,2) = 's'
        gfname(1,-1,3) = 'b'
        gfname(1,+1,1) = 'u'
        gfname(1,+1,2) = 'c'
        gfname(1,+1,3) = 't'
        dgfname(+1)    = ' '
        dgfname(-1)    = ' anti-'
        do 100 fs=1,nfs
        do 100 i=1,8
        fname(fs,i)  = gfname(quark(fs,i),iso(fs,i),gen(fs,i))
        dfname(fs,i) = dgfname(dir(i)*(-1)**i)
100     continue

C*** define particle polarizations
          pp     = 0
          pm     = 0
          pol(3) = 0
          pol(4) = 0
          pol(5) = 0
          pol(6) = 0
          pol(7) = 0
          pol(8) = 0

C*** define particle directions
        dir(1) =  1
        dir(2) =  1
        dir(3) = -1
        dir(4) = -1
        dir(5) = -1
        dir(6) = -1
        dir(7) = -1
        dir(8) = -1

	end

************************************************************************
        subroutine procout
************************************************************************
*	output for process specifications
*-----------------------------------------------------------------------
*       14.3.2001 Stefan Dittmaier
************************************************************************
        implicit real*8 (a-z)
        complex*16 cvf(20,-1:1,-3:3,8),cvif(20,-1:1,-3:3,8),csbt
        complex*16 cvfvf(20,-1:1,-3:3,8,-1:1,-3:3,8),mf(0:4)
        complex*16 cvvv(-3:3,-3:3,-3:3),cwwvv(-3:3,-3:3,-3:3,-3:3)
        complex*16 cvvs(-3:3,-3:3,-3:3)
        real*8 q(20,8),mp(-4:4),mp2(-4:4),gp(-4:4),polw(20,0:255)
	real*8 sym(20),wgt(20)
        real*8 colint(20,24,24),colintgg(20,24,24),colintg(20,24,24)
        integer fv(20,8,-3:3),qprop,quark(20,8),iso(20,8),gen(20,8)
        integer dir(8),pol(8),nfs,i,j,out,fs,fpermsgn(20,576)
        integer nfperm(20),fperm(20,576,8),qvperm(20,576,4)
        integer qlperm(20,576,0:4),ncclass(20),cclass(20,24,0:4)
        integer coldia(20),colav(20),qg
        character*6 fname(20,8),dfname(20,8)
	character*3 ndir(-1:1)

        common/lparam/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     &               mw,mw2,gw,mz,mz2,gz,mh,mh2,gh,me,me2,mt,mt2,gt
	common/out/out
        common/processpar/cvf,cvif,cvfvf,csbt,cwwvv,cvvv,cvvs,gs2,
     &          mp,mp2,gp,mf,q,pp,pm,polw,sym,wgt,
     &          fv,qprop,quark,iso,gen,dir,pol,nfs,fname,dfname
        common/generic/nfperm,fperm,qvperm,fpermsgn,qlperm
        common/colour/colint,colintgg,colintg,
     &          ncclass,cclass,coldia,colav
        common/qcdoption/qg

1000	format(/,'Process no.: ',i3,/,'============')
2000	format(/,'Fermions:',9x,'species',6x,'dir',5x,'pol',5x,
     &	       'Q',5x,'I_w',/,'---------')
3000	format(a6,'f(',i1,'): ',4x,2a6,3x,a3,4x,i5,2x,f5.2,2x,f5.2)
4000	format(a6,'f(',i1,'): ',4x,2a6,3x,a3,4x,i5,2x,f5.2,2x,f5.2)
7000	format(/,'Gauge-boson propagators:',3x,
     &		 'qprop =',i2,
     &	       /,'------------------------'
     &	       /,4x,'qprop = 0:   widths = 0',
     &	       /,4x,'        1:   constant width',
     &	       /,4x,'        2:   constant width * theta(p^2)',
     &	       /,4x,'        3:   running  width',
     &	       /,4x,'        4:   complex-mass scheme')
8000	format(/,'Colour structure:',/,'-----------------')
9000	format(/,'QCD background diagrams included (0=no/1=yes): ',
     &		  3x,i2)

	ndir(-1) = 'out'
	ndir(+1) = ' in'

	write(out,*)
	write(out,*)'                  \\          _/        '
	write(out,*)'               f2 _\\|        /| f3     '
	write(out,*)'                    \\       /          '
	write(out,*)'                     XXXXXXX           '
	write(out,*)'            f1 --<--XXXXXXXXX--<-- f4  '
	write(out,*)'                    XXXXXXXXX          '
	write(out,*)'                    XXXXXXXXX          '
	write(out,*)'            f8 -->--XXXXXXXXX-->-- f5  '
	write(out,*)'                     XXXXXXX           '
	write(out,*)'                    /       \\_        '
	write(out,*)'               f7 |/_       |\\  f6    '
	write(out,*)'                  /           \\       '

	do 150 fs=1,nfs
        write(out,1000) fs
	write(out,2000) 

	do i=1,2
	  write(out,3000) dfname(fs,i),i,dfname(fs,i),fname(fs,i),
     &	        ndir(dir(i)),pol(i),
     &		(-1)**i*dir(i)*q(fs,i),(-1)**i*dir(i)*iso(fs,i)/2d0
	enddo
	do i=3,8
	  write(out,4000) dfname(fs,i),i,dfname(fs,i),fname(fs,i),
     &	        ndir(dir(i)),pol(i),
     &		(-1)**i*dir(i)*q(fs,i),(-1)**i*dir(i)*iso(fs,i)/2d0
	enddo

200	continue

        write(out,8000) 
        write(out,*) 'colav   = ',colav(fs)
        write(out,*) 'coldia  = ',coldia(fs)
        write(out,*) 'ncclass = ',ncclass(fs)
        write(out,*) 'cclass  = ',(cclass(fs,j,0),j=1,ncclass(fs))
        write(out,*) 'colour interference matrix:'
        do i=1,ncclass(fs)
          write(out,*) (colint(fs,i,j),j=1,ncclass(fs))
        enddo
150	continue

	write(out,7000) qprop
	write(out,9000) qg

	end

